==> klayout-0.19.3.tar.gz: klayout-0.19.3/src/dbLibraryProxy.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbLibraryProxy.h"
#include "dbLibraryManager.h"
#include "dbLibrary.h"
#include "dbLayout.h"
#include "dbLayoutUtils.h"

namespace db
{

LibraryProxy::LibraryProxy (db::cell_index_type ci, db::Layout &layout, lib_id_type lib_id, cell_index_type cell_index)
  : Cell (ci, layout), m_lib_id (lib_id), m_cell_index (cell_index)
{
  // .. nothing yet ..
}

LibraryProxy::~LibraryProxy ()
{
  // .. nothing yet ..
}

Cell *
LibraryProxy::clone (Layout &layout) const
{
  return new LibraryProxy (cell_index (), layout, lib_id (), cell_index ());
}

std::vector<unsigned int> 
LibraryProxy::get_layer_indices (db::Layout &layout)
{
  std::vector<unsigned int> m_layer_indices; // @@@ should be somewhere "global" ..

  Library *lib = LibraryManager::instance ().lib (lib_id ());
  tl_assert (lib != 0);

  bool reuse_layer_list = (m_layer_indices.size () == lib->layout ().layers ());
  for (unsigned int i = 0; i < m_layer_indices.size () && reuse_layer_list; ++i) {
    reuse_layer_list = layout.is_valid_layer (m_layer_indices[i]) 
                         && lib->layout ().is_valid_layer (i)
                         && layout.get_properties (m_layer_indices[i]).log_equal (lib->layout ().get_properties (i));
  }

  if (! reuse_layer_list) {

    m_layer_indices.clear ();
    m_layer_indices.reserve (lib->layout ().layers ());

    for (unsigned int i = 0; i < lib->layout ().layers (); ++i) {

      if (! lib->layout ().is_valid_layer (i)) {
        m_layer_indices.push_back (i);
      } else {

        bool found = false;
        for (size_t l = 0; l < layout.layers () && !found; ++l) {
          if (layout.is_valid_layer (l) && layout.get_properties (l).log_equal (lib->layout ().get_properties (i))) {
            m_layer_indices.push_back (l);
            found = true;
          }
        }

        if (! found) {
          m_layer_indices.push_back (layout.insert_layer (lib->layout ().get_properties (i)));
        }

      }

    }

  }

  return m_layer_indices;
}

class LibraryCellIndexMapper
{
public:
  LibraryCellIndexMapper (Layout &layout, Library *lib)
    : mp_lib (lib), mp_layout (&layout)
  {
    // .. nothing yet ..
  }

  cell_index_type operator() (cell_index_type cell_index_in_lib)
  {
    return mp_layout->get_lib_proxy (mp_lib, cell_index_in_lib);
  }

private:
  Library *mp_lib;
  Layout *mp_layout;
};

void 
LibraryProxy::update ()
{
  std::vector<unsigned int> layer_mapping (get_layer_indices (layout ()));

  Library *lib = LibraryManager::instance ().lib (lib_id ());
  const db::Cell &source_cell = lib->layout ().cell (cell_index ());

  // @@@ TODO: DBU mapping!
  if (fabs (layout ().dbu () - lib->layout ().dbu ()) > 1e-6) {
    throw tl::Exception(tl::translate ("Library and layout must have the same database unit currently"));
  }

  clear_shapes ();
  clear_insts ();

  PropertyMapper prop_id_map (layout (), lib->layout ());

  for (unsigned int l = 0; l < lib->layout ().layers (); ++l) {
    if (lib->layout ().is_valid_layer (l)) {
      Shapes &target_shapes = shapes (layer_mapping [l]);
      for (db::Shapes::shape_iterator sh = source_cell.shapes (l).begin (db::Shapes::shape_iterator::All); ! sh.at_end (); ++sh) {
        target_shapes.insert (*sh, prop_id_map);
      }
    }
  }

  LibraryCellIndexMapper cell_index_mapper (layout (), lib);

  for (Cell::const_iterator inst = source_cell.begin (); !inst.at_end (); ++inst) {
    insert (*inst, cell_index_mapper, prop_id_map);
  }
}

std::string 
LibraryProxy::get_basic_name () const
{
  Library *lib = LibraryManager::instance ().lib (lib_id ());
  if (lib) {
    return lib->layout ().cell (cell_index ()).get_basic_name ();
  } else {
    return Cell::get_basic_name ();
  }
}

std::string 
LibraryProxy::get_display_name () const
{
  Library *lib = LibraryManager::instance ().lib (lib_id ());
  if (lib) {
    return lib->get_name () + "." + lib->layout ().cell (cell_index ()).get_display_name ();
  } else {
    return Cell::get_display_name ();
  }
}

}

==> klayout-0.21.19.tar.gz: klayout-0.21.19/src/dbLibraryProxy.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/


#include "dbLibraryProxy.h"
#include "dbLibraryManager.h"
#include "dbLibrary.h"
#include "dbLayout.h"
#include "dbLayoutUtils.h"

namespace db
{

LibraryProxy::LibraryProxy (db::cell_index_type ci, db::Layout &layout, lib_id_type lib_id, cell_index_type cell_index)
  : Cell (ci, layout), m_lib_id (lib_id), m_cell_index (cell_index)
{
  // .. nothing yet ..
}

LibraryProxy::~LibraryProxy ()
{
  // .. nothing yet ..
}

Cell *
LibraryProxy::clone (Layout &layout) const
{
  return new LibraryProxy (cell_index (), layout, lib_id (), cell_index ());
}

std::vector<unsigned int> 
LibraryProxy::get_layer_indices (db::Layout &layout)
{
  std::vector<unsigned int> m_layer_indices; // @@@ should be somewhere "global" ..

  Library *lib = LibraryManager::instance ().lib (lib_id ());
  tl_assert (lib != 0);

  bool reuse_layer_list = (m_layer_indices.size () == lib->layout ().layers ());
  for (unsigned int i = 0; i < m_layer_indices.size () && reuse_layer_list; ++i) {
    reuse_layer_list = layout.is_valid_layer (m_layer_indices[i]) 
                         && lib->layout ().is_valid_layer (i)
                         && layout.get_properties (m_layer_indices[i]).log_equal (lib->layout ().get_properties (i));
  }

  if (! reuse_layer_list) {

    m_layer_indices.clear ();
    m_layer_indices.reserve (lib->layout ().layers ());

    for (unsigned int i = 0; i < lib->layout ().layers (); ++i) {

      if (! lib->layout ().is_valid_layer (i)) {
        m_layer_indices.push_back (i);
      } else {

        bool found = false;
        for (unsigned int l = 0; l < layout.layers () && !found; ++l) {
          if (layout.is_valid_layer (l) && layout.get_properties (l).log_equal (lib->layout ().get_properties (i))) {
            m_layer_indices.push_back (l);
            found = true;
          }
        }

        if (! found) {
          m_layer_indices.push_back (layout.insert_layer (lib->layout ().get_properties (i)));
        }

      }

    }

  }

  return m_layer_indices;
}

class LibraryCellIndexMapper
{
public:
  LibraryCellIndexMapper (Layout &layout, Library *lib)
    : mp_lib (lib), mp_layout (&layout)
  {
    // .. nothing yet ..
  }

  cell_index_type operator() (cell_index_type cell_index_in_lib)
  {
    return mp_layout->get_lib_proxy (mp_lib, cell_index_in_lib);
  }

private:
  Library *mp_lib;
  Layout *mp_layout;
};

void 
LibraryProxy::update ()
{
  std::vector<unsigned int> layer_mapping (get_layer_indices (layout ()));

  Library *lib = LibraryManager::instance ().lib (lib_id ());
  const db::Cell &source_cell = lib->layout ().cell (cell_index ());

  // @@@ TODO: DBU mapping!
  if (fabs (layout ().dbu () - lib->layout ().dbu ()) > 1e-6) {
    throw tl::Exception(tl::translate ("Library and layout must have the same database unit currently"));
  }

  clear_shapes ();
  clear_insts ();

  PropertyMapper prop_id_map (layout (), lib->layout ());

  for (unsigned int l = 0; l < lib->layout ().layers (); ++l) {
    if (lib->layout ().is_valid_layer (l)) {
      Shapes &target_shapes = shapes (layer_mapping [l]);
      for (db::Shapes::shape_iterator sh = source_cell.shapes (l).begin (db::Shapes::shape_iterator::All); ! sh.at_end (); ++sh) {
        target_shapes.insert (*sh, prop_id_map);
      }
    }
  }

  LibraryCellIndexMapper cell_index_mapper (layout (), lib);

  for (Cell::const_iterator inst = source_cell.begin (); !inst.at_end (); ++inst) {
    insert (*inst, cell_index_mapper, prop_id_map);
  }
}

std::string 
LibraryProxy::get_basic_name () const
{
  Library *lib = LibraryManager::instance ().lib (lib_id ());
  if (lib) {
    return lib->layout ().cell (cell_index ()).get_basic_name ();
  } else {
    return Cell::get_basic_name ();
  }
}

std::string 
LibraryProxy::get_display_name () const
{
  Library *lib = LibraryManager::instance ().lib (lib_id ());
  if (lib) {
    return lib->get_name () + "." + lib->layout ().cell (cell_index ()).get_display_name ();
  } else {
    return Cell::get_display_name ();
  }
}

}
